# This version number is stored here instead of in `__init__.py` because
# namespace packages and the packages which live inside them can be imported
# in any order, which means their contents are supposed to be identical. If
# their contents differ, Python makes no guarantee about which one is loaded.
#
# See https://www.python.org/dev/peps/pep-0420/#namespace-packages-today

__version__ = '2.5.1'
